==> convokit-2.5.1.tar.gz: convokit-2.5.1/convokit/forecaster/CRAFT/CRAFTNN.py <==
try:
    import torch
except (ModuleNotFoundError, ImportError) as e:
    raise ModuleNotFoundError("torch is not currently installed. Run 'pip install convokit[craft]' if you would like to use the CRAFT model.")

from torch import nn
import os
from urllib.request import urlretrieve
from .CRAFTUtil import CONSTANTS

class EncoderRNN(nn.Module):
    """
    This module represents the utterance encoder component of CRAFT,
    responsible for creating vector representations of utterances
    """
    def __init__(self, hidden_size, embedding, n_layers=1, dropout=0):
        super(EncoderRNN, self).__init__()
        self.n_layers = n_layers
        self.hidden_size = hidden_size
        self.embedding = embedding

        # Initialize GRU; the input_size and hidden_size params are both set to 'hidden_size'
        #   because our input size is a word embedding with number of features == hidden_size
        self.gru = nn.GRU(hidden_size, hidden_size, n_layers,
                          dropout=(0 if n_layers == 1 else dropout), bidirectional=True)

    def forward(self, input_seq, input_lengths, hidden=None):
        # Convert word indexes to embeddings
        embedded = self.embedding(input_seq)
        # Pack padded batch of sequences for RNN module
        packed = torch.nn.utils.rnn.pack_padded_sequence(embedded, input_lengths)
        # Forward pass through GRU
        outputs, hidden = self.gru(packed, hidden)
        # Unpack padding
        outputs, _ = torch.nn.utils.rnn.pad_packed_sequence(outputs)
        # Sum bidirectional GRU outputs
        outputs = outputs[:, :, :self.hidden_size] + outputs[:, :, self.hidden_size:]
        # Return output and final hidden state
        return outputs, hidden

class ContextEncoderRNN(nn.Module):
    """This module represents the context encoder component of CRAFT, responsible for creating an order-sensitive vector representation of conversation context"""
    def __init__(self, hidden_size, n_layers=1, dropout=0):
        super(ContextEncoderRNN, self).__init__()
        self.n_layers = n_layers
        self.hidden_size = hidden_size

        # only unidirectional GRU for context encoding
        self.gru = nn.GRU(hidden_size, hidden_size, n_layers,
                          dropout=(0 if n_layers == 1 else dropout), bidirectional=False)

    def forward(self, input_seq, input_lengths, hidden=None):
        # Pack padded batch of sequences for RNN module
        packed = torch.nn.utils.rnn.pack_padded_sequence(input_seq, input_lengths)
        # Forward pass through GRU
        outputs, hidden = self.gru(packed, hidden)
        # Unpack padding
        outputs, _ = torch.nn.utils.rnn.pad_packed_sequence(outputs)
        # return output and final hidden state
        return outputs, hidden


class SingleTargetClf(nn.Module):
    """This module represents the CRAFT classifier head, which takes the context encoding and uses it to make a forecast"""
    def __init__(self, hidden_size, dropout=0.1):
        super(SingleTargetClf, self).__init__()

        self.hidden_size = hidden_size

        # initialize classifier
        self.layer1 = nn.Linear(hidden_size, hidden_size)
        self.layer1_act = nn.LeakyReLU()
        self.layer2 = nn.Linear(hidden_size, hidden_size // 2)
        self.layer2_act = nn.LeakyReLU()
        self.clf = nn.Linear(hidden_size // 2, 1)
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, encoder_outputs, encoder_input_lengths):
        # from stackoverflow (https://stackoverflow.com/questions/50856936/taking-the-last-state-from-bilstm-bigru-in-pytorch)
        # First we unsqueeze seqlengths two times so it has the same number of
        # of dimensions as output_forward
        # (batch_size) -> (1, batch_size, 1)
        lengths = encoder_input_lengths.unsqueeze(0).unsqueeze(2)
        # Then we expand it accordingly
        # (1, batch_size, 1) -> (1, batch_size, hidden_size)
        lengths = lengths.expand((1, -1, encoder_outputs.size(2)))

        # take only the last state of the encoder for each batch
        last_outputs = torch.gather(encoder_outputs, 0, lengths-1).squeeze(dim=0)
        # forward pass through hidden layers
        layer1_out = self.layer1_act(self.layer1(self.dropout(last_outputs)))
        layer2_out = self.layer2_act(self.layer2(self.dropout(layer1_out)))
        # compute and return logits
        logits = self.clf(self.dropout(layer2_out)).squeeze(dim=1)
        return logits


class Predictor(nn.Module):
    """This helper module encapsulates the CRAFT pipeline, defining the logic of passing an input through each consecutive sub-module."""
    def __init__(self, encoder, context_encoder, classifier):
        super(Predictor, self).__init__()
        self.encoder = encoder
        self.context_encoder = context_encoder
        self.classifier = classifier

    def forward(self, input_batch, dialog_lengths, dialog_lengths_list, utt_lengths, batch_indices, dialog_indices, batch_size, max_length):
        # Forward input through encoder model
        _, utt_encoder_hidden = self.encoder(input_batch, utt_lengths)

        # Convert utterance encoder final states to batched dialogs for use by context encoder
        context_encoder_input = makeContextEncoderInput(utt_encoder_hidden, dialog_lengths_list, batch_size, batch_indices, dialog_indices)

        # Forward pass through context encoder
        context_encoder_outputs, context_encoder_hidden = self.context_encoder(context_encoder_input, dialog_lengths)

        # Forward pass through classifier to get prediction logits
        logits = self.classifier(context_encoder_outputs, dialog_lengths)

        # Apply sigmoid activation
        predictions = torch.sigmoid(logits)
        return predictions

def makeContextEncoderInput(utt_encoder_hidden, dialog_lengths, batch_size, batch_indices, dialog_indices):
    """The utterance encoder takes in utterances in combined batches, with no knowledge of which ones go where in which conversation.
       Its output is therefore also unordered. We correct this by using the information computed during tensor conversion to regroup
       the utterance vectors into their proper conversational order."""
    # first, sum the forward and backward encoder states
    utt_encoder_summed = utt_encoder_hidden[-2,:,:] + utt_encoder_hidden[-1,:,:]
    # we now have hidden state of shape [utterance_batch_size, hidden_size]
    # split it into a list of [hidden_size,] x utterance_batch_size
    last_states = [t.squeeze() for t in utt_encoder_summed.split(1, dim=0)]

    # create a placeholder list of tensors to group the states by source dialog
    states_dialog_batched = [[None for _ in range(dialog_lengths[i])] for i in range(batch_size)]

    # group the states by source dialog
    for hidden_state, batch_idx, dialog_idx in zip(last_states, batch_indices, dialog_indices):
        states_dialog_batched[batch_idx][dialog_idx] = hidden_state

    # stack each dialog into a tensor of shape [dialog_length, hidden_size]
    states_dialog_batched = [torch.stack(d) for d in states_dialog_batched]

    # finally, condense all the dialog tensors into a single zero-padded tensor
    # of shape [max_dialog_length, batch_size, hidden_size]
    return torch.nn.utils.rnn.pad_sequence(states_dialog_batched)

def initialize_model(custom_model_path, voc, device, device_type: str, hidden_size, encoder_n_layers, dropout,
                     context_encoder_n_layers):
    print("Loading saved parameters...")
    if custom_model_path is None:
        if not os.path.isfile("model.tar"):
            print("\tDownloading trained CRAFT...")
            urlretrieve(CONSTANTS['MODEL_URL'], "model.tar")
            print("\t...Done!")
        custom_model_path = "model.tar"
    # If running in a non-GPU environment, you need to tell PyTorch to convert the parameters to CPU tensor format.
    # To do so, replace the previous line with the following:
    if device_type == 'cpu':
        checkpoint = torch.load(custom_model_path, map_location=torch.device('cpu'))
    elif device_type == 'cuda':
        checkpoint = torch.load(custom_model_path)
    encoder_sd = checkpoint['en']
    context_sd = checkpoint['ctx']
    if 'atk_clf' in checkpoint:
        attack_clf_sd = checkpoint['atk_clf']

    embedding_sd = checkpoint['embedding']
    voc.__dict__ = checkpoint['voc_dict']

    print('Building encoders, decoder, and classifier...')
    # Initialize word embeddings
    embedding = nn.Embedding(voc.num_words, hidden_size)
    embedding.load_state_dict(embedding_sd)
    # Initialize utterance and context encoders
    encoder: EncoderRNN = EncoderRNN(hidden_size, embedding, encoder_n_layers, dropout)
    context_encoder: ContextEncoderRNN = ContextEncoderRNN(hidden_size, context_encoder_n_layers, dropout)
    encoder.load_state_dict(encoder_sd)
    context_encoder.load_state_dict(context_sd)
    # Initialize classifier
    attack_clf: SingleTargetClf = SingleTargetClf(hidden_size, dropout)
    if 'atk_clf' in checkpoint:
        attack_clf.load_state_dict(attack_clf_sd)
    # Use appropriate device
    encoder = encoder.to(device)
    context_encoder = context_encoder.to(device)
    attack_clf = attack_clf.to(device)
    print('Models built and ready to go!')

    # Set dropout layers to eval mode
    encoder.eval()
    context_encoder.eval()
    attack_clf.eval()

    # Initialize the pipeline
    return Predictor(encoder, context_encoder, attack_clf)

==> convokit-3.0.0.tar.gz: convokit-3.0.0/convokit/forecaster/CRAFT/CRAFTNN.py <==
try:
    import torch
except (ModuleNotFoundError, ImportError) as e:
    raise ModuleNotFoundError(
        "torch is not currently installed. Run 'pip install convokit[craft]' if you would like to use the CRAFT model."
    )

from torch import nn
import os
from urllib.request import urlretrieve
from .CRAFTUtil import CONSTANTS


class EncoderRNN(nn.Module):
    """
    This module represents the utterance encoder component of CRAFT,
    responsible for creating vector representations of utterances
    """

    def __init__(self, hidden_size, embedding, n_layers=1, dropout=0):
        super(EncoderRNN, self).__init__()
        self.n_layers = n_layers
        self.hidden_size = hidden_size
        self.embedding = embedding

        # Initialize GRU; the input_size and hidden_size params are both set to 'hidden_size'
        #   because our input size is a word embedding with number of features == hidden_size
        self.gru = nn.GRU(
            hidden_size,
            hidden_size,
            n_layers,
            dropout=(0 if n_layers == 1 else dropout),
            bidirectional=True,
        )

    def forward(self, input_seq, input_lengths, hidden=None):
        # Convert word indexes to embeddings
        embedded = self.embedding(input_seq)
        # Pack padded batch of sequences for RNN module
        packed = torch.nn.utils.rnn.pack_padded_sequence(embedded, input_lengths)
        # Forward pass through GRU
        outputs, hidden = self.gru(packed, hidden)
        # Unpack padding
        outputs, _ = torch.nn.utils.rnn.pad_packed_sequence(outputs)
        # Sum bidirectional GRU outputs
        outputs = outputs[:, :, : self.hidden_size] + outputs[:, :, self.hidden_size :]
        # Return output and final hidden state
        return outputs, hidden


class ContextEncoderRNN(nn.Module):
    """This module represents the context encoder component of CRAFT, responsible for creating an order-sensitive vector representation of conversation context"""

    def __init__(self, hidden_size, n_layers=1, dropout=0):
        super(ContextEncoderRNN, self).__init__()
        self.n_layers = n_layers
        self.hidden_size = hidden_size

        # only unidirectional GRU for context encoding
        self.gru = nn.GRU(
            hidden_size,
            hidden_size,
            n_layers,
            dropout=(0 if n_layers == 1 else dropout),
            bidirectional=False,
        )

    def forward(self, input_seq, input_lengths, hidden=None):
        # Pack padded batch of sequences for RNN module
        packed = torch.nn.utils.rnn.pack_padded_sequence(input_seq, input_lengths)
        # Forward pass through GRU
        outputs, hidden = self.gru(packed, hidden)
        # Unpack padding
        outputs, _ = torch.nn.utils.rnn.pad_packed_sequence(outputs)
        # return output and final hidden state
        return outputs, hidden


class SingleTargetClf(nn.Module):
    """This module represents the CRAFT classifier head, which takes the context encoding and uses it to make a forecast"""

    def __init__(self, hidden_size, dropout=0.1):
        super(SingleTargetClf, self).__init__()

        self.hidden_size = hidden_size

        # initialize classifier
        self.layer1 = nn.Linear(hidden_size, hidden_size)
        self.layer1_act = nn.LeakyReLU()
        self.layer2 = nn.Linear(hidden_size, hidden_size // 2)
        self.layer2_act = nn.LeakyReLU()
        self.clf = nn.Linear(hidden_size // 2, 1)
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, encoder_outputs, encoder_input_lengths):
        # from stackoverflow (https://stackoverflow.com/questions/50856936/taking-the-last-state-from-bilstm-bigru-in-pytorch)
        # First we unsqueeze seqlengths two times so it has the same number of
        # of dimensions as output_forward
        # (batch_size) -> (1, batch_size, 1)
        lengths = encoder_input_lengths.unsqueeze(0).unsqueeze(2)
        # Then we expand it accordingly
        # (1, batch_size, 1) -> (1, batch_size, hidden_size)
        lengths = lengths.expand((1, -1, encoder_outputs.size(2)))

        # take only the last state of the encoder for each batch
        last_outputs = torch.gather(encoder_outputs, 0, lengths - 1).squeeze(dim=0)
        # forward pass through hidden layers
        layer1_out = self.layer1_act(self.layer1(self.dropout(last_outputs)))
        layer2_out = self.layer2_act(self.layer2(self.dropout(layer1_out)))
        # compute and return logits
        logits = self.clf(self.dropout(layer2_out)).squeeze(dim=1)
        return logits


class Predictor(nn.Module):
    """This helper module encapsulates the CRAFT pipeline, defining the logic of passing an input through each consecutive sub-module."""

    def __init__(self, encoder, context_encoder, classifier):
        super(Predictor, self).__init__()
        self.encoder = encoder
        self.context_encoder = context_encoder
        self.classifier = classifier

    def forward(
        self,
        input_batch,
        dialog_lengths,
        dialog_lengths_list,
        utt_lengths,
        batch_indices,
        dialog_indices,
        batch_size,
        max_length,
    ):
        # Forward input through encoder model
        _, utt_encoder_hidden = self.encoder(input_batch, utt_lengths)

        # Convert utterance encoder final states to batched dialogs for use by context encoder
        context_encoder_input = makeContextEncoderInput(
            utt_encoder_hidden, dialog_lengths_list, batch_size, batch_indices, dialog_indices
        )

        # Forward pass through context encoder
        context_encoder_outputs, context_encoder_hidden = self.context_encoder(
            context_encoder_input, dialog_lengths
        )

        # Forward pass through classifier to get prediction logits
        logits = self.classifier(context_encoder_outputs, dialog_lengths)

        # Apply sigmoid activation
        predictions = torch.sigmoid(logits)
        return predictions


def makeContextEncoderInput(
    utt_encoder_hidden, dialog_lengths, batch_size, batch_indices, dialog_indices
):
    """The utterance encoder takes in utterances in combined batches, with no knowledge of which ones go where in which conversation.
    Its output is therefore also unordered. We correct this by using the information computed during tensor conversion to regroup
    the utterance vectors into their proper conversational order."""
    # first, sum the forward and backward encoder states
    utt_encoder_summed = utt_encoder_hidden[-2, :, :] + utt_encoder_hidden[-1, :, :]
    # we now have hidden state of shape [utterance_batch_size, hidden_size]
    # split it into a list of [hidden_size,] x utterance_batch_size
    last_states = [t.squeeze() for t in utt_encoder_summed.split(1, dim=0)]

    # create a placeholder list of tensors to group the states by source dialog
    states_dialog_batched = [[None for _ in range(dialog_lengths[i])] for i in range(batch_size)]

    # group the states by source dialog
    for hidden_state, batch_idx, dialog_idx in zip(last_states, batch_indices, dialog_indices):
        states_dialog_batched[batch_idx][dialog_idx] = hidden_state

    # stack each dialog into a tensor of shape [dialog_length, hidden_size]
    states_dialog_batched = [torch.stack(d) for d in states_dialog_batched]

    # finally, condense all the dialog tensors into a single zero-padded tensor
    # of shape [max_dialog_length, batch_size, hidden_size]
    return torch.nn.utils.rnn.pad_sequence(states_dialog_batched)


def initialize_model(
    custom_model_path,
    voc,
    device,
    device_type: str,
    hidden_size,
    encoder_n_layers,
    dropout,
    context_encoder_n_layers,
):
    print("Loading saved parameters...")
    if custom_model_path is None:
        if not os.path.isfile("model.tar"):
            print("\tDownloading trained CRAFT...")
            urlretrieve(CONSTANTS["MODEL_URL"], "model.tar")
            print("\t...Done!")
        custom_model_path = "model.tar"
    # If running in a non-GPU environment, you need to tell PyTorch to convert the parameters to CPU tensor format.
    # To do so, replace the previous line with the following:
    if device_type == "cpu":
        checkpoint = torch.load(custom_model_path, map_location=torch.device("cpu"))
    elif device_type == "cuda":
        checkpoint = torch.load(custom_model_path)
    encoder_sd = checkpoint["en"]
    context_sd = checkpoint["ctx"]
    if "atk_clf" in checkpoint:
        attack_clf_sd = checkpoint["atk_clf"]

    embedding_sd = checkpoint["embedding"]
    voc.__dict__ = checkpoint["voc_dict"]

    print("Building encoders, decoder, and classifier...")
    # Initialize word embeddings
    embedding = nn.Embedding(voc.num_words, hidden_size)
    embedding.load_state_dict(embedding_sd)
    # Initialize utterance and context encoders
    encoder: EncoderRNN = EncoderRNN(hidden_size, embedding, encoder_n_layers, dropout)
    context_encoder: ContextEncoderRNN = ContextEncoderRNN(
        hidden_size, context_encoder_n_layers, dropout
    )
    encoder.load_state_dict(encoder_sd)
    context_encoder.load_state_dict(context_sd)
    # Initialize classifier
    attack_clf: SingleTargetClf = SingleTargetClf(hidden_size, dropout)
    if "atk_clf" in checkpoint:
        attack_clf.load_state_dict(attack_clf_sd)
    # Use appropriate device
    encoder = encoder.to(device)
    context_encoder = context_encoder.to(device)
    attack_clf = attack_clf.to(device)
    print("Models built and ready to go!")

    # Set dropout layers to eval mode
    encoder.eval()
    context_encoder.eval()
    attack_clf.eval()

    # Initialize the pipeline
    return Predictor(encoder, context_encoder, attack_clf)
